/**
 * Copyright 2009 Ibrahim Chaehoi
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * 	http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language governing permissions
 * and limitations under the License.
 */
package net.jawr.web.taglib;

import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.jawr.web.config.JawrConfig;
import net.jawr.web.exception.ResourceNotFoundException;
import net.jawr.web.resource.ImageResourcesHandler;
import net.jawr.web.resource.bundle.CheckSumUtils;
import net.jawr.web.resource.bundle.factory.util.PathNormalizer;
import net.jawr.web.servlet.RendererRequestUtils;

import org.apache.log4j.Logger;

/**
 * Utility class for image tags.
 * 
 * @author ibrahim Chaehoi
 */
public final class ImageTagUtils {

	/** The logger */
	private static Logger logger = Logger.getLogger(ImageTagUtils.class);
	
	/**
	 * Returns the image URL generated by Jawr from a source image path
	 * @param imgSrc the source image path
     * @param imgRsHandler the image resource handler
	 * @param request the request
	 * @param response the response
	 * @return the image URL generated by Jawr from a source image path
	 */
	public static String getImageUrl(String imgSrc, ImageResourcesHandler imgRsHandler, HttpServletRequest request, HttpServletResponse response) {
		
		if(imgRsHandler == null){
			throw new IllegalStateException("You are using a Jawr image tag while the Jawr Image servlet has not been initialized. Initialization of Jawr Image servlet either failed or never occurred.");
		}
	
		String contextPath = request.getContextPath();
		
		// relative path
		if(!imgSrc.startsWith("/")){ 
			imgSrc = PathNormalizer.concatWebPath(request.getRequestURI(), imgSrc);
			int idx = imgSrc.indexOf(contextPath);
			if(idx > -1){
				imgSrc = imgSrc.substring(idx+contextPath.length());
			}
		}
		
		String newUrl = (String) imgRsHandler.getCacheUrl(imgSrc);
		
        JawrConfig jawrConfig = imgRsHandler.getJawrConfig();
		if(newUrl == null){
        	try {
				newUrl = CheckSumUtils.getCacheBustedUrl(imgSrc, imgRsHandler.getRsHandler(), jawrConfig);
				imgRsHandler.addMapping(imgSrc, newUrl);
        	} catch (IOException e) {
	    		logger.info("Unable to create the checksum for the image '"+imgSrc+"' while generating image tag.");
			} catch (ResourceNotFoundException e) {
				logger.info("Unable to find the image '"+imgSrc+"' while generating image tag.");
			}
    	}
        
        if(newUrl == null){
        	newUrl = imgSrc;
        }
        
        String imageServletMapping = jawrConfig.getServletMapping();
		if("".equals(imageServletMapping)){
			if(newUrl.startsWith("/")){
				newUrl = newUrl.substring(1);
			}
		}else{
			newUrl = PathNormalizer.joinDomainToPath(imageServletMapping, newUrl);
		}
		
		boolean sslRequest = RendererRequestUtils.isSslRequest(request);
		
		newUrl = RendererRequestUtils.getRenderedUrl(newUrl, jawrConfig, contextPath, sslRequest);
		
		return response.encodeURL(newUrl);
	}

	

}
